using System;
using MeshVR;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.EventSystems;

namespace CheesyFX
{
    public class ClickListener : MonoBehaviour, IPointerClickHandler, IPointerEnterHandler, IPointerExitHandler
    {
        public readonly UnityEvent onLeftClick = new UnityEvent();
        public readonly UnityEvent onMiddleClick = new UnityEvent();
        public readonly UnityEvent onRightClick = new UnityEvent();
        public readonly UnityEvent onDragUp = new UnityEvent();
        public readonly UnityEvent onDragDown = new UnityEvent();
        public readonly UnityEvent onDragRight = new UnityEvent();
        public readonly UnityEvent onDragLeft = new UnityEvent();
        public readonly UnityEvent onPointerEnter = new UnityEvent();
        public readonly UnityEvent onPointerExit = new UnityEvent();

        private bool hovered;
        public bool dragEnabled;
        private bool leftMouseDown;
        private Vector2 mousePosition;

        public void OnPointerEnter(PointerEventData eventData)
        {
            hovered = true;
            onPointerEnter.Invoke();
        }
        
        public void OnPointerExit(PointerEventData eventData)
        {
            hovered = false;
            onPointerExit.Invoke();
            if (leftMouseDown && Input.GetKey(KeyCode.Mouse0))
            {
                Vector2 drag = (Vector2)Input.mousePosition - mousePosition;
                if (Mathf.Abs(drag.y) > Mathf.Abs(drag.x))
                {
                    if(drag.y > 0) onDragUp.Invoke();
                    else onDragDown.Invoke();
                }
                else
                {
                    if(drag.x > 0) onDragRight.Invoke();
                    else onDragLeft.Invoke();
                }
            }
        }
        
        public void OnPointerClick(PointerEventData eventData)
        {
            if (eventData.button == PointerEventData.InputButton.Left)
            {
                onLeftClick.Invoke();
            }
            else if (eventData.button == PointerEventData.InputButton.Middle)
            {
                onMiddleClick.Invoke();
            }
            else if (eventData.button == PointerEventData.InputButton.Right)
            {
                onRightClick.Invoke();
            }
        }

        private void Update()
        {
            if(!dragEnabled || !hovered) return;
            if (Input.GetKeyDown(KeyCode.Mouse0))
            {
                mousePosition = Input.mousePosition;
                leftMouseDown = true;
            }
        
            if (Input.GetKeyUp(KeyCode.Mouse0)) leftMouseDown = false;
        }
    }
}